using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using Leap.Unity;
using System.Diagnostics;
using GPUTools.Hair.Scripts.Runtime.Render;
using Battlehub.RTSaveLoad.UnityEngineNS.AINS;
using GPUTools.Physics.Scripts.Types.Dynamic;

namespace Stopper {
    [DefaultExecutionOrder(1000)]
	public class CUAClothing : MVRScript {
        Transform originalParent;

        Atom atom;
        bool connected = false;
        
        Atom cuaAtom;
        string cuaAtomUid;

        DAZBone parentBone;

        DAZCharacterSelector selector;
        PhysicsMaterialControl physics;
        DAZCharacterRun characterRun;

        DAZClothingItem dazClothingItem;
        DAZDynamic dazDynamic;
        DAZMesh dazMesh;
        DAZSkinWrap dazSkinWrap;

        
        DAZSkinV2 skin;

        ClothSettings clothSettings;

        DAZHairGroup dazHair;
        HairSimControl hairSimControl;
        DAZBone head;
        int numStrands;

        

        ComputeBuffer verticesBuffer;
        ComputeBuffer particlesBuffer;
        ComputeBuffer normalsBuffer;

        List<int> triangleIndexes = new List<int>();
        int minVertex, maxVertex;
        int[] triangles;
        Vector2[] uv;

        AsyncGPUReadbackRequest verticesRequest;
        AsyncGPUReadbackRequest normalsRequest;
        AsyncGPUReadbackRequest particlesRequest;

        
        JSONStorableUrl assetUrlJSON;
        JSONStorableString assetName;
        JSONStorableStringChooser assetChooser;
        JSONStorableStringChooser cuaAssetChooser;

        UIDynamicPopup assetChooserPopup;


        JSONStorableFloat vertexIndex;

        JSONStorableFloat xOffset;
        JSONStorableFloat yOffset;
        JSONStorableFloat zOffset;

        JSONStorableFloat xRotation;
        JSONStorableFloat yRotation;
        JSONStorableFloat zRotation;

        JSONStorableBool zeroOnReattach;

        Vector3 previousRotation;

        UIDynamicSlider vertexSlider;
        JSONStorableStringChooser parentBoneChooser;
        UIDynamicPopup parentBoneChooserPopup;
        UIDynamicPopup hairParticleTypePopup;

        List<UIDynamic> vertexAttachComponents = new List<UIDynamic>();
        List<UIDynamic> offsetSliders = new List<UIDynamic>();

        public class ConnectionType {
            public static ConnectionType Skin = new ConnectionType();
            public static ConnectionType Clothing = new ConnectionType();
            public static ConnectionType Bone = new ConnectionType();
            public static ConnectionType Hair = new ConnectionType();
        }

        ConnectionType connectionType = ConnectionType.Skin;

        public class HairParticleType {
            public static HairParticleType Physics = new HairParticleType();
            public static HairParticleType Render = new HairParticleType();
        }
        HairParticleType hairParticleType = HairParticleType.Physics;
        
		public override void Init() {
            WalkAndGetParentItems();

            cuaAtomUid = $"{containingAtom.name}#{dazDynamic.name}#{this.name}";

            // If we're loading from scene, the atom might already be present
            cuaAtom = SuperController.singleton.GetAtomByUid(cuaAtomUid);

            physics = containingAtom.GetStorableByID("PhysicsModel") as PhysicsMaterialControl;
            if(physics == null) { SuperController.LogError("Unable to find PhysicsModel. Is this a Person atom?"); return;}
            
            selector = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
            if(selector == null) { SuperController.LogError("Unable to find character selector. Is this a Person atom?"); return;}
            characterRun = selector.GetComponent<DAZCharacterRun>();

            head = containingAtom.GetStorableByID("head") as DAZBone;
            
            var selectFileButton = CreateButton("Select File", false);
            var clearButton = CreateButton("Clear", true);

            assetUrlJSON = new JSONStorableUrl("assetUrl", string.Empty, new JSONStorableString.SetStringCallback(this.SyncAssetUrl), "assetbundle|scene", "Custom\\Assets");
            assetUrlJSON.allowFullComputerBrowse = true;
            assetUrlJSON.allowBrowseAboveSuggestedPath = true;
            assetUrlJSON.beginBrowseWithObjectCallback = new JSONStorableUrl.BeginBrowseWithObjectCallback(this.BeginBrowse);
            assetUrlJSON.suggestedPathGroup = "UnityAsset";

            assetUrlJSON.fileBrowseButton = selectFileButton.button;
            assetUrlJSON.clearButton = clearButton.button;
            RegisterUrl(this.assetUrlJSON);


            assetName = new JSONStorableString("assetName", null);
            RegisterString(assetName);
            assetName.setCallbackFunction = (name) => {
                if(cuaAssetChooser != null) {
                    cuaAssetChooser.val = name;
                }
            };

            assetChooser = new JSONStorableStringChooser(
            "Asset", 
            new List<string>(),
            null, 
            "Asset", 
            (string asset) => {
                assetName.val = asset;
            });
            assetChooserPopup = CreateFilterablePopup(assetChooser, false);
            assetChooserPopup.gameObject.SetActive(false);
			

            JSONStorableStringChooser connectionTypeChooser = new JSONStorableStringChooser(
                "ConnectionType", 
                GetConnectionTypes(),
                "None", 
                "Connection Type", 
                (string type) => {
                    if(type == "Skin") {
                        connectionType = ConnectionType.Skin;
                        
                        skin = characterRun.skin;
                        StartCoroutine(ConnectSkin());
                        UIVertexAttachMode();
                    } else if (type == "Clothing") {
                        connectionType = ConnectionType.Clothing;
                        UIVertexAttachMode();
                        StartCoroutine(ConnectClothing());
                    } else if (type == "Bone") {
                        connectionType = ConnectionType.Bone;
                        connected = true;
                        UIBoneAttachMode();
                    } else if (type == "Hair") {
                        connectionType = ConnectionType.Hair;
                        UIVertexAttachMode(true);
                        StartCoroutine(ConnectHair());
                    } else {
                        Disconnect();
                    }
                });
            RegisterStringChooser(connectionTypeChooser);
            var itemTypePopup = CreateScrollablePopup(connectionTypeChooser, false);

            parentBoneChooser = new JSONStorableStringChooser(
                "Parent Bone", 
                GetBones(),
                null, 
                "Parent Bone", 
                (string boneName) => {
                    Disconnect();
                    parentBone = physics.GetComponentsInChildren<DAZBone>(true).First((bone) => bone.name == boneName);
                    connected = true;
                }
            );
		    parentBoneChooserPopup = CreateFilterablePopup(parentBoneChooser, false);
            parentBoneChooserPopup.popup.onOpenPopupHandlers += () => {
                parentBoneChooser.choices = GetBones();
			};
            RegisterStringChooser(parentBoneChooser);

             JSONStorableStringChooser hairParticleTypeChooser = new JSONStorableStringChooser(
                "HairParticleType", 
                new List<string>(){"Physics", "Render"},
                "Physics", 
                "Hair Particle Type", 
                (string type) => {
                    if(type == "Physics") {
                        hairParticleType = HairParticleType.Physics;
                    } else if(type == "Render") {
                        hairParticleType = HairParticleType.Render;
                    }
                    StartCoroutine(ConnectHair());
                });
            RegisterStringChooser(hairParticleTypeChooser);
            hairParticleTypePopup = CreateScrollablePopup(hairParticleTypeChooser, false);
            vertexAttachComponents.Add(hairParticleTypePopup);


            vertexIndex = new JSONStorableFloat("Vertex Index", 0, 0, 0, false, true);
            
            vertexIndex.storeType = JSONStorableParam.StoreType.Full;
            vertexSlider = this.CreateSlider(vertexIndex, false);
            vertexSlider.rangeAdjustEnabled = false;
            vertexSlider.slider.wholeNumbers = true;
            RegisterFloat(vertexIndex);
            vertexAttachComponents.Add(vertexSlider);


            vertexIndex.setCallbackFunction = (val) => {
                if(val > vertexIndex.max || val < 0) {
                    return;
                }

                minVertex = (int)vertexIndex.max;
                maxVertex = 0;
                if(triangles != null) {
                    triangleIndexes.Clear();
                    
                    for(int i = 0; i < triangles.Length; i+=3) {
                        for(int j = 0; j < 3; j++){ 
                            if(triangles[i+j] == (int)val) {
                                if(!triangleIndexes.Contains(i)) {
                                    triangleIndexes.Add(i);
                                }

                                for(int k = 0; k < 3; k++){ 
                                    minVertex = Math.Min(minVertex, triangles[i+k]);
                                    maxVertex = Math.Max(maxVertex, triangles[i+k]);
                                }
                            }
                        }
                    }

                    if(minVertex >= maxVertex) {
                        minVertex = 0;
                        maxVertex = (int)vertexIndex.max;
                    }
                } else {
                    minVertex = 0;
                    maxVertex = (int)vertexIndex.max;
                }
            };

            zeroOnReattach = new JSONStorableBool("zeroLocalOnReattach", false);
            var zeroOnReattachToggle = CreateToggle(zeroOnReattach, true);
            zeroOnReattachToggle.label = "Zero Offset On Reattach";
            vertexAttachComponents.Add(zeroOnReattachToggle);

            var attachToClosest = CreateButton("Re-attach To Closest", true);
            attachToClosest.button.onClick.AddListener(() => AttachToClosest());
            vertexAttachComponents.Add(attachToClosest);

            
            xOffset = new JSONStorableFloat("xOffset",0,-1,1,false);
            RegisterFloat(xOffset);
            offsetSliders.Add(this.CreateSlider(xOffset, false));
            xOffset.setCallbackFunction = (val) => SyncLocalPosition();

            yOffset = new JSONStorableFloat("yOffset",0,-1,1,false);
            RegisterFloat(yOffset);
            offsetSliders.Add(this.CreateSlider(yOffset, false));
            yOffset.setCallbackFunction = (val) => SyncLocalPosition();

            zOffset = new JSONStorableFloat("zOffset",0,-1,1,false);
            RegisterFloat(zOffset);
            offsetSliders.Add(this.CreateSlider(zOffset, false));
            zOffset.setCallbackFunction = (val) => SyncLocalPosition();

            var rotationWarning = CreateTextField(new JSONStorableString("rotationWarning", "Note: modifying rotation sliders directly might yield unpredictable behavior"), false);
            rotationWarning.height = 10;
            offsetSliders.Add(rotationWarning);
            xRotation = new JSONStorableFloat("xRotation",0,0,360,true);
            RegisterFloat(xRotation);
            offsetSliders.Add(this.CreateSlider(xRotation, false));
            xRotation.setCallbackFunction = (val) => SyncLocalPosition();

            yRotation = new JSONStorableFloat("yRotation",0,-0,360,true);
            RegisterFloat(yRotation);
            offsetSliders.Add(this.CreateSlider(yRotation, false));
            yRotation.setCallbackFunction = (val) => SyncLocalPosition();

            zRotation = new JSONStorableFloat("zRotation",0,-0,360,true);
            RegisterFloat(zRotation);
            offsetSliders.Add(this.CreateSlider(zRotation, false));
            zRotation.setCallbackFunction = (val) => SyncLocalPosition();

            
		}

        public void Start() {
        }

        private void UINoMode() {
            foreach(var uiDynamic in vertexAttachComponents) {
                uiDynamic.gameObject.SetActive(false);
            }
            foreach(var uiDynamic in offsetSliders) {
                uiDynamic.gameObject.SetActive(false);
            }
            parentBoneChooserPopup.gameObject.SetActive(false);
        }

        private void UIVertexAttachMode(bool hair = false) {
            UINoMode();
            foreach(var uiDynamic in vertexAttachComponents) {
                uiDynamic.gameObject.SetActive(true);
            }
            foreach(var uiDynamic in offsetSliders) {
                uiDynamic.gameObject.SetActive(true);
            }
            hairParticleTypePopup.gameObject.SetActive(hair);
        }

        private void UIBoneAttachMode() {
            UINoMode();
            foreach(var uiDynamic in offsetSliders) {
                uiDynamic.gameObject.SetActive(true);
            }
            parentBoneChooserPopup.gameObject.SetActive(true);
        }

        private void WalkAndGetParentItems(){
			GameObject parent = this.gameObject;
			while(containingAtom == null || parent != containingAtom.gameObject){
				if(parent.transform.parent == null){
					return;
				}
				
				if(dazClothingItem == null){
					dazClothingItem = parent.GetComponent<DAZClothingItem>();
				}

                if(dazHair == null){
					dazHair = parent.GetComponent<DAZHairGroup>();
				}
                

                if(dazMesh == null){
					dazMesh = parent.GetComponent<DAZMesh>();
				}

                if(dazDynamic == null){
					dazDynamic = parent.GetComponent<DAZDynamic>();
				}

				if(dazSkinWrap == null){
					dazSkinWrap = parent.GetComponent<DAZSkinWrap>();
				}

                containingAtom = parent.GetComponent<Atom>();
				parent = parent.transform.parent.gameObject;
			}
		}

        private List<string> GetConnectionTypes() {
            List<string> connectionTypes = new List<string>(){"None", "Skin", "Bone"};

            if(dazClothingItem != null) {
                connectionTypes.Add("Clothing");
            }

            if(dazHair != null) {
                connectionTypes.Add("Hair");
            }

            return connectionTypes;
        }

        private List<string> GetBones() {
            List<string> boneNames = new List<string>();
            foreach(var bone in physics.GetComponentsInChildren<DAZBone>(true)){
                boneNames.Add(bone.name);
            }
            return boneNames;
        }

        protected void BeginBrowse(JSONStorableUrl jsurl) {
            List<ShortCut> shortCutsForDirectory = FileManagerSecure.GetShortCutsForDirectory("Custom/Assets", true, true, true, true);
            jsurl.shortCuts = shortCutsForDirectory;
        }

        private IEnumerator CreateCUA(string url) {
            if(cuaAtom == null) {
                GameObject cuaPrefab = SuperController.singleton.GetCachedPrefab("a_mis", "CustomUnityAsset");
                if(cuaPrefab == null){
                    AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_mis", "CustomUnityAsset", typeof(GameObject));
                    if (request == null)
                    {
                        SuperController.LogError("Unable to load CUA prefab");
                        yield break;
                    }
                    yield return base.StartCoroutine(request);

                    cuaPrefab = request.GetAsset<GameObject>();
                    if (cuaPrefab != null)
                    {
                        SuperController.singleton.RegisterPrefab("a_mis", "CustomUnityAsset", cuaPrefab);
                    } else {
                        SuperController.LogError("Unable to load CUA prefab");
                        yield break;
                    }
                } 
                var cua = Instantiate(cuaPrefab);
                cuaAtom = cua.GetComponent<Atom>();
                atom = containingAtom;
                SuperController.singleton.AddAtom(cuaAtom, cuaAtomUid, false, false, false, false);
            }

            cuaAtom.hidden = true;
            SyncLocalPosition();
            
            CustomUnityAssetLoader cuaLoader = cuaAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>();
            
            var urlJson = cuaLoader.GetUrlJSONParam("assetUrl");
            urlJson.val = url;

            cuaAssetChooser = cuaLoader.GetStringChooserJSONParam("assetName");
            
            yield return new WaitUntil (() => cuaAssetChooser.choices.Count > 1);
            
            assetChooserPopup.gameObject.SetActive(true);
            assetChooser.choices = new List<string>(cuaAssetChooser.choices);

            if(assetName.val != null) {
                assetChooser.val = assetName.val;
                assetName.setCallbackFunction(assetName.val);
            }
        }

        protected void SyncAssetUrl(string url) {
            StartCoroutine(CreateCUA(url));
        }

        private List<string> GetAtomUids() {
            List<string> atomUids = new List<string>(){"None"};
            atomUids.AddRange(SuperController.singleton.GetAtoms().Where(a => a.type == "Person").Select(a => a.uid));
            return atomUids;
        }

        private List<string> GetItemUids() {
            List<string> itemUids = new List<string>(){"None"};

            if(atom != null) {
                if(connectionType == ConnectionType.Clothing) {
                    itemUids.AddRange(atom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList());
                } else if(connectionType == ConnectionType.Skin) {
                    itemUids.Add("Skin");
                } 
            }

            return itemUids;
        }

        private void Disconnect() {
            connected = false;
            verticesBuffer = null;
            normalsBuffer = null;
            triangles = null;
            uv = null;
        }

        private IEnumerator ConnectSkin() {
            Disconnect();
            
            yield return SwapSkin();
            
            connected = true;
       
            yield break;
        }

        private IEnumerator SwapSkin() {
            skin = characterRun.skin;
            if(!skin.isActiveAndEnabled) {
                yield return new WaitUntil(() => skin.isActiveAndEnabled);
            }

            yield return null;
            SyncBuffers();
            
            var dazMesh = skin.dazMesh;
            triangles = dazMesh.UVTriangles;
            uv = dazMesh.UV;
            
            vertexIndex.max = verticesBuffer.count-1;
            vertexIndex.setCallbackFunction(vertexIndex.val);
        }

        private IEnumerator ConnectClothing() {
            Disconnect();

            
            if(!dazClothingItem.enabled && dazClothingItem.ready)
                yield return new WaitUntil(() => dazClothingItem.enabled && dazClothingItem.ready);

            dazSkinWrap = dazClothingItem.GetComponentInChildren<DAZSkinWrap>();
            if(dazSkinWrap == null || !dazSkinWrap.enabled) {
                yield return new WaitUntil(() => {
                    dazSkinWrap = dazClothingItem.GetComponentInChildren<DAZSkinWrap>();
                    return dazSkinWrap != null && dazSkinWrap.enabled;
                });
            }

            yield return null;
            
            clothSettings = dazClothingItem.GetComponentInChildren<ClothSettings>();
            SyncBuffers();

            var dazMesh = dazSkinWrap.dazMesh;
            triangles = dazMesh.UVTriangles;
            uv = dazMesh.UV;
            
            vertexIndex.max = verticesBuffer.count-1;
            vertexIndex.setCallbackFunction(vertexIndex.val);
          
            connected = true;

        
            yield break;
        }

        private IEnumerator ConnectHair() {
            Disconnect();

            if(!dazHair.enabled)
                yield return new WaitUntil(() => dazHair.enabled);

            hairSimControl = dazHair.GetComponentInChildren<HairSimControl>();
            if(hairSimControl == null || !hairSimControl.enabled){
                yield return new WaitUntil(() => {
                    hairSimControl = dazHair.GetComponentInChildren<HairSimControl>();
                    return hairSimControl != null && hairSimControl.enabled;
                });
            }
            yield return null;
            SyncBuffers();
            vertexIndex.max = particlesBuffer.count-2;
            
            vertexIndex.setCallbackFunction(vertexIndex.val);
            connected = true;

            yield break;
        }

        private void AttachToClosest() {
            if(!connected) {
                return;
            }
            
            float closestDistance = float.MaxValue;
            int closestVertex = 0;
            Vector3 controllerPosition = cuaAtom.mainController.transform.position;

            if(connectionType == ConnectionType.Hair) {
                if(hairParticleType == HairParticleType.Physics) {
                    GPParticle[] particles  = new GPParticle[particlesBuffer.count];
                    particlesBuffer.GetData(particles);
                    for(int i = 0; i < particles.Length; i++) {
                        float sqrDistance = (controllerPosition - particles[i].Position).sqrMagnitude;
                        if(sqrDistance < closestDistance) {
                            closestVertex = i;
                            closestDistance = sqrDistance;
                        }
                    }
                } else if(hairParticleType == HairParticleType.Render) {
                    TessRenderParticle[] particles  = new TessRenderParticle[particlesBuffer.count];
                    particlesBuffer.GetData(particles);
                    for(int i = 0; i < particles.Length; i++) {
                        float sqrDistance = (controllerPosition - particles[i].Position).sqrMagnitude;
                        if(sqrDistance < closestDistance) {
                            closestVertex = i;
                            closestDistance = sqrDistance;
                        }
                    }
                }
            } else {
                Vector3[] vertices = new Vector3[verticesBuffer.count];
                verticesBuffer.GetData(vertices);
                for(int i = 0; i < vertices.Length; i++) {
                    float sqrDistance = (controllerPosition - vertices[i]).sqrMagnitude;
                    if(sqrDistance < closestDistance) {
                        closestVertex = i;
                        closestDistance = sqrDistance;
                    }
                }
            }

            vertexIndex.val = closestVertex;

            if(zeroOnReattach.val) {
                xRotation.valNoCallback = 0;
                yRotation.valNoCallback = 0;
                zRotation.valNoCallback = 0;

                xOffset.valNoCallback = 0;
                yOffset.valNoCallback = 0;
                zOffset.valNoCallback = 0;
                SyncLocalPosition();
            }            
        }

        private void SyncBuffers() {
            if(connectionType == ConnectionType.Clothing) {
                if(clothSettings != null && clothSettings.enabled) {
                    verticesBuffer = clothSettings.Runtime.ClothOnlyVertices.ComputeBuffer;
                } else {
                    verticesBuffer = dazSkinWrap.PreCalculatedVerticesBuffer.ComputeBuffer;
                    // Need to call dispatch to copy verts into the pre-compute buffer
                    // Perf hit should be minor
                    dazSkinWrap.Dispatch();
                }
                normalsBuffer = dazSkinWrap.NormalsBuffer.ComputeBuffer;
            } else if (connectionType == ConnectionType.Skin) {
                if(skin != characterRun.skin) {
                    StartCoroutine(SwapSkin());
                }

                if (skin.delayDisplayOneFrame) {
                    verticesBuffer = skin.delayedVertsBuffer;
                } else if (skin.useSmoothing) {
                    verticesBuffer = skin.smoothedVertsBuffer;
                } else {
                    verticesBuffer = skin.rawVertsBuffer;
                }
                normalsBuffer = skin.NormalsBuffer.ComputeBuffer;
          
            } else if (connectionType == ConnectionType.Hair) {
                if(hairParticleType == HairParticleType.Physics) {
                    particlesBuffer = hairSimControl.hairSettings.RuntimeData.Particles.ComputeBuffer;
                } else if(hairParticleType == HairParticleType.Render) {
                    particlesBuffer = hairSimControl.hairSettings.RuntimeData.TessRenderParticles.ComputeBuffer;
                }
                numStrands = hairSimControl.hairSettings.StandsSettings.Provider.GetStandsNum();
            }
        }

        private void SyncLocalPosition() {
            if(cuaAtom != null) {
                cuaAtom.mainController.transform.localPosition = new Vector3(xOffset.val, yOffset.val, zOffset.val);
                cuaAtom.mainController.transform.localEulerAngles = new Vector3(xRotation.val, yRotation.val, zRotation.val);
            }
        }

        private bool isConnected() {
            if(!connected)
                return false;

            if(cuaAtom == null || !cuaAtom.enabled)
                return false;                

            if(connectionType == ConnectionType.Clothing) {
                if(dazSkinWrap == null) return false;
                if(!dazSkinWrap.enabled) return false;
                if(vertexIndex.val > vertexIndex.max || vertexIndex.val < 0) return false;
            } else if(connectionType == ConnectionType.Skin) {
                if(skin == null) return false;
                if(!skin.enabled) return false;
                if(vertexIndex.val > vertexIndex.max || vertexIndex.val < 0) return false;
            } else if(connectionType == ConnectionType.Hair) {
                if(hairSimControl == null) return false;
                if(!hairSimControl.enabled) return false;
            }

            return true;        
        }

        public void LateUpdate() {
            
            if(!isConnected()) {
                return;
            }

            var rotation = cuaAtom.mainController.transform.localEulerAngles;
            if(rotation != previousRotation) {
                xRotation.valNoCallback = rotation.x;
                yRotation.valNoCallback = rotation.y;
                zRotation.valNoCallback = rotation.z;
            }
            previousRotation = rotation;

            xOffset.valNoCallback = cuaAtom.mainController.transform.localPosition.x;
            yOffset.valNoCallback = cuaAtom.mainController.transform.localPosition.y;
            zOffset.valNoCallback = cuaAtom.mainController.transform.localPosition.z;
            
            if(connectionType != ConnectionType.Bone){
                SyncBuffers();
            }
            
            if(connectionType == ConnectionType.Bone) {
                if(parentBone != null) {
                    cuaAtom.reParentObject.position = parentBone.transform.position;
                    cuaAtom.reParentObject.rotation = parentBone.transform.rotation;
                }
            } else if(connectionType == ConnectionType.Hair) {
                int index = (int)vertexIndex.val;
                if(particlesBuffer != null) {
                    if(index >= particlesBuffer.count) {
                        return;
                    }
                    if(hairParticleType == HairParticleType.Physics) {
                        particlesRequest = AsyncGPUReadback.Request(particlesBuffer, GPParticle.Size()*2, index*GPParticle.Size());
                    } else if(hairParticleType == HairParticleType.Render) {
                        particlesRequest = AsyncGPUReadback.Request(particlesBuffer, TessRenderParticle.Size()*2, index*TessRenderParticle.Size());
                    }
                } else {
                    return;
                }
                
                if(!particlesRequest.done) {
                    particlesRequest.WaitForCompletion();
                }
                
                if(particlesRequest.hasError) {
                    return;
                }

                var particlePosition = Vector3.zero;
                var nextParticlePosition = Vector3.zero;
               
                if(hairParticleType == HairParticleType.Physics) {
                    var particles = particlesRequest.GetData<GPParticle>();
                    particlePosition = particles[0].Position;
                    nextParticlePosition = particles[1].Position;

                } else if(hairParticleType == HairParticleType.Render) {
                    var particles = particlesRequest.GetData<TessRenderParticle>();
                    particlePosition = particles[0].Position;
                    nextParticlePosition = particles[1].Position;
                }

                if(particlePosition.ContainsNaN() || nextParticlePosition.ContainsNaN()){
                    Log("nan");
                    return;
                } 

                Vector3 normal;

                if((index +1) % numStrands == 0) { //next particle is in another strand
                    normal = Vector3.down;
                } else {
                    normal = (nextParticlePosition - particlePosition).normalized;
                }
                Vector3 tangent = (head.transform.position - particlePosition).normalized;
                
                cuaAtom.reParentObject.position = particlePosition;
                cuaAtom.reParentObject.rotation = Quaternion.LookRotation(normal, tangent);

            } else {
                int vertex = (int)vertexIndex.val;
                if(verticesBuffer != null && normalsBuffer != null) {
                    if(vertex >= verticesBuffer.count) {
                        return;
                    }
                    verticesRequest = AsyncGPUReadback.Request(verticesBuffer, (maxVertex-minVertex+1)*12, minVertex*12);
                    normalsRequest = AsyncGPUReadback.Request(normalsBuffer, 12, vertex*12);
                } else {
                    return;
                }

                if(!verticesRequest.done) {
                    verticesRequest.WaitForCompletion();
                }
                if(!normalsRequest.done) {
                    normalsRequest.WaitForCompletion();
                }

                if(verticesRequest.hasError || normalsRequest.hasError) {
                    return;
                }

                var vertices = verticesRequest.GetData<Vector3>();
                var normals = normalsRequest.GetData<Vector3>();


                Vector3 position = vertices[vertex-minVertex];
                Vector3 normal = normals[0];

                if(position.ContainsNaN() || normal.ContainsNaN()){
                    return;
                } 
                
                Vector3 tangent = GetTangent(vertex, triangleIndexes, minVertex, triangles, vertices, normals, uv);

                cuaAtom.reParentObject.position = position;
                cuaAtom.reParentObject.rotation = Quaternion.LookRotation(normal, tangent);
            }
        }

        public static void Log( string str) {
            if(UnityEngine.Time.frameCount % 10 == 0) {
                SuperController.LogMessage(str);
            }
        }

        public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.valueFormat = "F4";
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return storableFloat;
        }

        // Originally Copied from https://discussions.unity.com/t/calculating-tangents-vector4/179
        public Vector3 GetTangent(int vertex, List<int> triangleIndexes, int vertexOffset, int[] triangles, NativeArray<Vector3> vertices, NativeArray<Vector3> normals, Vector2[] uv) {
            Vector3 tan = new Vector3();

            foreach (int a in triangleIndexes) {
                int i1 = triangles[a + 0];
                int i2 = triangles[a + 1];
                int i3 = triangles[a + 2];

                Vector3 v1 = vertices[i1-vertexOffset];
                Vector3 v2 = vertices[i2-vertexOffset];
                Vector3 v3 = vertices[i3-vertexOffset];

                Vector2 w1 = uv[i1];
                Vector2 w2 = uv[i2];
                Vector2 w3 = uv[i3];

                float x1 = v2.x - v1.x;
                float x2 = v3.x - v1.x;
                float y1 = v2.y - v1.y;
                float y2 = v3.y - v1.y;
                float z1 = v2.z - v1.z;
                float z2 = v3.z - v1.z;

                float s1 = w2.x - w1.x;
                float s2 = w3.x - w1.x;
                float t1 = w2.y - w1.y;
                float t2 = w3.y - w1.y;

                float r = 1.0f / (s1 * t2 - s2 * t1);

                Vector3 sdir = new Vector3((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r, (t2 * z1 - t1 * z2) * r);

                if(i1 == vertex)
                    tan += sdir;

                if(i2 == vertex)
                    tan += sdir;

                if(i3 == vertex)
                    tan += sdir;;

            }

            
            Vector3 n = normals[0];
            Vector3 t = tan;

            Vector3 tmp = (t - n * Vector3.Dot(n, t)).normalized;
            return new Vector3(tmp.x, tmp.y, tmp.z);

        }

        private void OnEnable() {
            if(cuaAtom != null) {
                cuaAtom.gameObject.SetActive(true);
            }
        }

        private void OnDisable() {
            if(cuaAtom != null) {
                cuaAtom.gameObject.SetActive(false);
            }
        }

        private void OnDestroy() {
            if(cuaAtom != null) {
                SuperController.singleton.RemoveAtom(cuaAtom);
            }
        }
	}
}